@@ -1,28 +1,31 @@
+"""State space models that classify trajectories as well as decode the
+trajectory from population spiking
+"""
 from copy import deepcopy
 from logging import getLogger
 
 import joblib
 import matplotlib.pyplot as plt
 import numpy as np
+import seaborn as sns
 import sklearn
 import xarray as xr
-from replay_trajectory_classification.bins import atleast_2d, get_centers
 from replay_trajectory_classification.continuous_state_transitions import (
     EmpiricalMovement, RandomWalk, Uniform)
 from replay_trajectory_classification.core import (_acausal_classify,
                                                    _acausal_classify_gpu,
                                                    _causal_classify,
-                                                   _causal_classify_gpu, mask,
+                                                   _causal_classify_gpu,
+                                                   atleast_2d, check_converged,
+                                                   get_centers, mask,
                                                    scaled_likelihood)
-from replay_trajectory_classification.discrete_state_transitions import \
-    DiagonalDiscrete
+from replay_trajectory_classification.discrete_state_transitions import (
+    DiagonalDiscrete, estimate_discrete_state_transition)
 from replay_trajectory_classification.environments import Environment
 from replay_trajectory_classification.initial_conditions import \
     UniformInitialConditions
 from replay_trajectory_classification.likelihoods import (
-    _ClUSTERLESS_ALGORITHMS, estimate_place_fields,
-    estimate_spiking_likelihood)
-from replay_trajectory_classification.misc import NumbaKDE
+    _SORTED_SPIKES_ALGORITHMS, _ClUSTERLESS_ALGORITHMS)
 from replay_trajectory_classification.observation_model import ObservationModel
 from sklearn.base import BaseEstimator
 
@@ -31,10 +34,14 @@ logger = getLogger(__name__)
 sklearn.set_config(print_changed_only=False)
 
 _DEFAULT_CLUSTERLESS_MODEL_KWARGS = {
-    'model': NumbaKDE,
-    'model_kwargs': {
-        'bandwidth': np.array([24.0, 24.0, 24.0, 24.0, 6.0, 6.0])
-    }
+    'mark_std': 24.0,
+    'position_std': 6.0,
+}
+
+_DEFAULT_SORTED_SPIKES_MODEL_KWARGS = {
+    'position_std': 6.0,
+    'use_diffusion_distance': False,
+    'block_size': None,
 }
 
 _DEFAULT_CONTINUOUS_TRANSITIONS = (
@@ -68,7 +75,7 @@ class _ClassifierBase(BaseEstimator):
     def fit_environments(self, position, environment_labels=None):
         for environment in self.environments:
             if environment_labels is None:
-                is_environment = np.ones((position.shape[0],), dtype=np.bool)
+                is_environment = np.ones((position.shape[0],), dtype=bool)
             else:
                 is_environment = (environment_labels ==
                                   environment.environment_name)
@@ -102,11 +109,11 @@ class _ClassifierBase(BaseEstimator):
             encoding_group_labels=None,
             environment_labels=None,
     ):
-        logger.info('Fitting state transition...')
+        logger.info('Fitting continuous state transition...')
 
         if is_training is None:
             n_time = position.shape[0]
-            is_training = np.ones((n_time,), dtype=np.bool)
+            is_training = np.ones((n_time,), dtype=bool)
 
         if encoding_group_labels is None:
             n_time = position.shape[0]
@@ -139,21 +146,130 @@ class _ClassifierBase(BaseEstimator):
                     row_ind, column_ind, :st.shape[0], :st.shape[1]] = st
 
     def fit_discrete_state_transition(self):
+        logger.info('Fitting discrete state transition')
         n_states = len(self.continuous_transition_types)
         self.discrete_state_transition_ = (
             self.discrete_transition_type.make_state_transition(n_states))
 
-    def _get_results(self, likelihood, n_time, time, state_names, use_gpu,
-                     is_compute_acausal):
+    def plot_discrete_state_transition(
+            self, state_names=None, cmap='Oranges', ax=None,
+            convert_to_seconds=False, sampling_frequency=1):
+
+        if ax is None:
+            ax = plt.gca()
+
+        if state_names is None:
+            state_names = [f'state {ind + 1}' for ind in
+                           range(self.discrete_state_transition_.shape[0])]
+
+        if convert_to_seconds:
+            discrete_state_transition = (
+                1 / (1 - self.discrete_state_transition_)) / sampling_frequency
+            vmin, vmax, fmt = 0.0, None, '0.03f'
+            label = 'Seconds'
+        else:
+            discrete_state_transition = self.discrete_state_transition_
+            vmin, vmax, fmt = 0.0, 1.0, '0.03f'
+            label = 'Probability'
+
+        sns.heatmap(data=discrete_state_transition,
+                    vmin=vmin, vmax=vmax, annot=True, fmt=fmt, cmap=cmap,
+                    xticklabels=state_names, yticklabels=state_names, ax=ax,
+                    cbar_kws={'label': label})
+        ax.set_ylabel('Previous State', fontsize=12)
+        ax.set_xlabel('Current State', fontsize=12)
+        ax.set_title('Discrete State Transition', fontsize=16)
+
+    def estimate_parameters(self, fit_args, predict_args, tolerance=1E-4,
+                            max_iter=10, verbose=True, store_likelihood=True,
+                            estimate_initial_conditions=True,
+                            estimate_discrete_transition=True):
+
+        if 'store_likelihood' in predict_args:
+            store_likelihood = predict_args['store_likelihood']
+        else:
+            predict_args['store_likelihood'] = store_likelihood
+
+        self.fit(**fit_args)
+        results = self.predict(**predict_args)
+
+        data_log_likelihoods = [results.data_log_likelihood]
+        log_likelihood_change = np.inf
+        converged = False
+        increasing = True
+        n_iter = 0
+        n_time = len(results.time)
+
+        logger.info(
+            f'iteration {n_iter}, likelihood: {data_log_likelihoods[-1]}')
+        get_results_args = {
+            key: value for key, value in predict_args.items()
+            if key in ['time', 'state_names', 'use_gpu', 'is_compute_acausal']
+        }
+
+        while not converged and (n_iter < max_iter):
+            if estimate_initial_conditions:
+                self.initial_conditions_ = results.isel(
+                    time=0).acausal_posterior.values[..., np.newaxis]
+            if estimate_discrete_transition:
+                self.discrete_state_transition_ = estimate_discrete_state_transition(
+                    self, results)
+
+            if store_likelihood:
+                results = self._get_results(
+                    self.likelihood_,
+                    n_time,
+                    **get_results_args)
+            else:
+                results = self.predict(**predict_args)
+
+            data_log_likelihoods.append(results.data_log_likelihood)
+            log_likelihood_change = (
+                data_log_likelihoods[-1] - data_log_likelihoods[-2])
+            n_iter += 1
+
+            converged, increasing = check_converged(
+                data_log_likelihoods[-1], data_log_likelihoods[-2], tolerance)
+
+            if verbose:
+                logger.info(
+                    f'iteration {n_iter}, '
+                    f'likelihood: {data_log_likelihoods[-1]}, '
+                    f'change: {log_likelihood_change}'
+                )
+
+        if not converged and (n_iter == max_iter):
+            logger.warning('Max iterations reached...')
+
+        return results, data_log_likelihoods
+
+    def _get_results(
+        self,
+        likelihood,
+        n_time,
+        time=None,
+        is_compute_acausal=True,
+        use_gpu=False,
+        state_names=None
+    ):
         n_states = self.discrete_state_transition_.shape[0]
         results = {}
-        results['likelihood'] = np.full(
-            (n_time, n_states, self.max_pos_bins_, 1), np.nan)
+
+        if use_gpu:
+            results['likelihood'] = np.full(
+                (n_time, n_states, self.max_pos_bins_, 1), np.nan,
+                dtype=np.float32)
+        else:
+            results['likelihood'] = np.full(
+                (n_time, n_states, self.max_pos_bins_, 1), np.nan,
+                dtype=np.float64)
+
         for state_ind, obs in enumerate(self.observation_models):
             likelihood_name = (obs.environment_name, obs.encoding_group)
             n_bins = likelihood[likelihood_name].shape[1]
             results['likelihood'][:, state_ind, :n_bins] = (
                 likelihood[likelihood_name][..., np.newaxis])
+
         results['likelihood'] = scaled_likelihood(
             results['likelihood'], axis=(1, 2))
         results['likelihood'][np.isnan(results['likelihood'])] = 0.0
@@ -174,10 +290,12 @@ class _ClassifierBase(BaseEstimator):
                 (results['causal_posterior'][:, :, is_track_interior],
                  data_log_likelihood) = (
                     _causal_classify(
-                        self.initial_conditions_[:, is_track_interior],
-                        self.continuous_state_transition_[st_interior_ind],
-                        self.discrete_state_transition_,
-                        results['likelihood'][:, :, is_track_interior]))
+                        self.initial_conditions_[
+                            :, is_track_interior].astype(np.float64),
+                        self.continuous_state_transition_[
+                            st_interior_ind].astype(np.float64),
+                        self.discrete_state_transition_.astype(np.float64),
+                        results['likelihood'][:, :, is_track_interior].astype(np.float64)))
             else:
                 results['causal_posterior'] = np.full(
                     (n_time, n_states, n_position_bins, 1), np.nan,
@@ -185,33 +303,37 @@ class _ClassifierBase(BaseEstimator):
                 (results['causal_posterior'][:, :, is_track_interior],
                  data_log_likelihood) = (
                     _causal_classify_gpu(
-                        self.initial_conditions_[:, is_track_interior],
-                        self.continuous_state_transition_[st_interior_ind],
-                        self.discrete_state_transition_,
-                        results['likelihood'][:, :, is_track_interior]))
+                        self.initial_conditions_[
+                            :, is_track_interior].astype(np.float32),
+                        self.continuous_state_transition_[
+                            st_interior_ind].astype(np.float32),
+                        self.discrete_state_transition_.astype(np.float32),
+                        results['likelihood'][:, :, is_track_interior].astype(np.float32)))
 
             if is_compute_acausal:
                 logger.info('Estimating acausal posterior...')
                 if not use_gpu:
                     results['acausal_posterior'] = np.full(
                         (n_time, n_states, n_position_bins, 1), np.nan,
-                        dtype=np.float64)
+                        dtype=np.float32)
                     results['acausal_posterior'][:, :, is_track_interior] = (
                         _acausal_classify(
-                            results['causal_posterior'][:,
-                                                        :, is_track_interior],
-                            self.continuous_state_transition_[st_interior_ind],
-                            self.discrete_state_transition_))
+                            results['causal_posterior'][:, :,
+                                                        is_track_interior].astype(np.float64),
+                            self.continuous_state_transition_[
+                                st_interior_ind].astype(np.float64),
+                            self.discrete_state_transition_.astype(np.float64)))
                 else:
                     results['acausal_posterior'] = np.full(
                         (n_time, n_states, n_position_bins, 1), np.nan,
                         dtype=np.float32)
                     results['acausal_posterior'][:, :, is_track_interior] = (
                         _acausal_classify_gpu(
-                            results['causal_posterior'][:,
-                                                        :, is_track_interior],
-                            self.continuous_state_transition_[st_interior_ind],
-                            self.discrete_state_transition_))
+                            results['causal_posterior'][:, :,
+                                                        is_track_interior].astype(np.float32),
+                            self.continuous_state_transition_[
+                                st_interior_ind].astype(np.float32),
+                            self.discrete_state_transition_.astype(np.float32)))
 
             if time is None:
                 time = np.arange(n_time)
@@ -288,7 +410,7 @@ class _ClassifierBase(BaseEstimator):
             results = xr.Dataset(
                 {key: (dims,
                        (mask(value, is_track_interior).squeeze(axis=-1)
-                        .reshape(new_shape).swapaxes(-1, -2)))
+                        .reshape(new_shape, order='F')))
                  for key, value in results.items()},
                 coords=coords,
                 attrs=attrs)
@@ -376,7 +498,7 @@ class _ClassifierBase(BaseEstimator):
 
 
 class SortedSpikesClassifier(_ClassifierBase):
-    '''
+    """
 
     Attributes
     ----------
@@ -390,30 +512,32 @@ class SortedSpikesClassifier(_ClassifierBase):
         The initial conditions class instance to fit
     infer_track_interior : bool, optional
         Whether to infer the valid position bins
-    knot_spacing : float, optional
-        How far apart the spline knots are in position
-    spike_model_penalty : float, optional
-        L2 penalty (ridge) for the size of the regression coefficients
+    sorted_spikes_algorithm : str
+        The type of algorithm. See _SORTED_SPIKES_ALGORITHMS for keys
+    sorted_spikes_algorithm_params : dict
+        Parameters for the algorithm.
 
-    '''
+    """
 
-    def __init__(self,
-                 environments=_DEFAULT_ENVIRONMENT,
-                 observation_models=None,
-                 continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
-                 discrete_transition_type=DiagonalDiscrete(0.98),
-                 initial_conditions_type=UniformInitialConditions(),
-                 infer_track_interior=True,
-                 knot_spacing=10,
-                 spike_model_penalty=1E1):
+    def __init__(
+        self,
+        environments=_DEFAULT_ENVIRONMENT,
+        observation_models=None,
+        continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
+        discrete_transition_type=DiagonalDiscrete(0.98),
+        initial_conditions_type=UniformInitialConditions(),
+        infer_track_interior=True,
+        sorted_spikes_algorithm='spiking_likelihood_kde',
+        sorted_spikes_algorithm_params=_DEFAULT_SORTED_SPIKES_MODEL_KWARGS,
+    ):
         super().__init__(environments,
                          observation_models,
                          continuous_transition_types,
                          discrete_transition_type,
                          initial_conditions_type,
                          infer_track_interior)
-        self.knot_spacing = knot_spacing
-        self.spike_model_penalty = spike_model_penalty
+        self.sorted_spikes_algorithm = sorted_spikes_algorithm
+        self.sorted_spikes_algorithm_params = sorted_spikes_algorithm_params
 
     def fit_place_fields(self,
                          position,
@@ -424,7 +548,7 @@ class SortedSpikesClassifier(_ClassifierBase):
         logger.info('Fitting place fields...')
         n_time = position.shape[0]
         if is_training is None:
-            is_training = np.ones((n_time,), dtype=np.bool)
+            is_training = np.ones((n_time,), dtype=bool)
 
         if encoding_group_labels is None:
             encoding_group_labels = np.zeros((n_time,), dtype=np.int32)
@@ -435,6 +559,10 @@ class SortedSpikesClassifier(_ClassifierBase):
 
         is_training = np.asarray(is_training).squeeze()
 
+        kwargs = self.sorted_spikes_algorithm_params
+        if kwargs is None:
+            kwargs = {}
+
         self.place_fields_ = {}
         for obs in np.unique(self.observation_models):
             environment = self.environments[
@@ -444,13 +572,16 @@ class SortedSpikesClassifier(_ClassifierBase):
             is_environment = (environment_labels == obs.environment_name)
             likelihood_name = (obs.environment_name, obs.encoding_group)
 
-            self.place_fields_[likelihood_name] = estimate_place_fields(
+            self.place_fields_[likelihood_name] = _SORTED_SPIKES_ALGORITHMS[
+                self.sorted_spikes_algorithm][0](
                 position=position[is_training & is_encoding & is_environment],
                 spikes=spikes[is_training & is_encoding & is_environment],
                 place_bin_centers=environment.place_bin_centers_,
                 place_bin_edges=environment.place_bin_edges_,
-                penalty=self.spike_model_penalty,
-                knot_spacing=self.knot_spacing)
+                edges=environment.edges_,
+                is_track_interior=environment.is_track_interior_,
+                is_track_boundary=environment.is_track_boundary_,
+                **kwargs)
 
     def plot_place_fields(self, sampling_frequency=1, figsize=(10, 7)):
         try:
@@ -485,24 +616,25 @@ class SortedSpikesClassifier(_ClassifierBase):
             encoding_group_labels=None,
             environment_labels=None,
             ):
-        '''
+        """Fit the spatial grid, initial conditions, place field model, and
+        transition matrices.
 
         Parameters
         ----------
-        position : ndarray, shape (n_time, n_position_dims)
-        spikes : ndarray, shape (n_time, n_neurons)
-        is_training : None or bool ndarray, shape (n_time), optional
+        position : np.ndarray, shape (n_time, n_position_dims)
+        spikes : np.ndarray, shape (n_time, n_neurons)
+        is_training : None or bool np.ndarray, shape (n_time), optional
             Time bins to be used for encoding.
-        encoding_group_labels : None or ndarray, shape (n_time,)
+        encoding_group_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding encoding group for each time point
-        environment_labels : None or ndarray, shape (n_time,)
+        environment_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding environment for each time point
 
         Returns
         -------
         self
 
-        '''
+        """
         position = atleast_2d(np.asarray(position))
         spikes = np.asarray(spikes)
         self.fit_environments(position, environment_labels)
@@ -525,12 +657,13 @@ class SortedSpikesClassifier(_ClassifierBase):
 
     def predict(self, spikes, time=None, is_compute_acausal=True,
                 use_gpu=False,
-                state_names=None):
-        '''
+                state_names=None,
+                store_likelihood=False):
+        """Run the state space model on spikes.
 
         Parameters
         ----------
-        spikes : ndarray, shape (n_time, n_neurons)
+        spikes : np.ndarray, shape (n_time, n_neurons)
         time : ndarray or None, shape (n_time,), optional
         is_compute_acausal : bool, optional
         use_gpu : bool, optional
@@ -541,7 +674,7 @@ class SortedSpikesClassifier(_ClassifierBase):
         -------
         results : xarray.Dataset
 
-        '''
+        """
         spikes = np.asarray(spikes)
         n_time = spikes.shape[0]
 
@@ -552,17 +685,20 @@ class SortedSpikesClassifier(_ClassifierBase):
             env_ind = self.environments.index(env_name)
             is_track_interior = (
                 self.environments[env_ind].is_track_interior_.ravel(order='F'))
-            likelihood[(env_name, enc_group)] = estimate_spiking_likelihood(
+            likelihood[(env_name, enc_group)] = _SORTED_SPIKES_ALGORITHMS[
+                self.sorted_spikes_algorithm][1](
                 spikes,
                 place_fields.values,
                 is_track_interior)
+        if store_likelihood:
+            self.likelihood_ = likelihood
 
         return self._get_results(
-            likelihood, n_time, time, state_names, use_gpu, is_compute_acausal)
+            likelihood, n_time, time, is_compute_acausal, use_gpu, state_names)
 
 
 class ClusterlessClassifier(_ClassifierBase):
-    '''
+    """
 
     Attributes
     ----------
@@ -581,7 +717,7 @@ class ClusterlessClassifier(_ClassifierBase):
     clusterless_algorithm_params : dict
         Parameters for the clusterless algorithms.
 
-    '''
+    """
 
     def __init__(self,
                  environments=_DEFAULT_ENVIRONMENT,
@@ -609,23 +745,23 @@ class ClusterlessClassifier(_ClassifierBase):
                        is_training=None,
                        encoding_group_labels=None,
                        environment_labels=None):
-        '''
+        """Fit the clusterless place field model.
 
         Parameters
         ----------
         position : array_like, shape (n_time, n_position_dims)
         multiunits : array_like, shape (n_time, n_marks, n_electrodes)
         is_training : None or array_like, shape (n_time,)
-        encoding_group_labels : None or ndarray, shape (n_time,)
+        encoding_group_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding encoding group for each time point
-        environment_labels : None or ndarray, shape (n_time,)
+        environment_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding environment for each time point
 
-        '''
+        """
         logger.info('Fitting multiunits...')
         n_time = position.shape[0]
         if is_training is None:
-            is_training = np.ones((n_time,), dtype=np.bool)
+            is_training = np.ones((n_time,), dtype=bool)
 
         if encoding_group_labels is None:
             encoding_group_labels = np.zeros((n_time,), dtype=np.int32)
@@ -657,8 +793,9 @@ class ClusterlessClassifier(_ClassifierBase):
                     position=position[is_group],
                     multiunits=multiunits[is_group],
                     place_bin_centers=environment.place_bin_centers_,
-                    is_track_interior=environment.is_track_interior_.ravel(
-                        order='F'),
+                    is_track_interior=environment.is_track_interior_,
+                    is_track_boundary=environment.is_track_boundary_,
+                    edges=environment.edges_,
                     **kwargs
             )
 
@@ -669,23 +806,24 @@ class ClusterlessClassifier(_ClassifierBase):
             encoding_group_labels=None,
             environment_labels=None,
             ):
-        '''
+        """Fit the spatial grid, initial conditions, place field model, and
+        transition matrices.
 
         Parameters
         ----------
         position : array_like, shape (n_time, n_position_dims)
         multiunits : array_like, shape (n_time, n_marks, n_electrodes)
         is_training : None or array_like, shape (n_time,)
-        encoding_group_labels : None or ndarray, shape (n_time,)
+        encoding_group_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding encoding group for each time point
-        environment_labels : None or ndarray, shape (n_time,)
+        environment_labels : None or np.ndarray, shape (n_time,)
             Label for the corresponding environment for each time point
 
         Returns
         -------
         self
 
-        '''
+        """
         position = atleast_2d(np.asarray(position))
         multiunits = np.asarray(multiunits)
 
@@ -708,24 +846,26 @@ class ClusterlessClassifier(_ClassifierBase):
         return self
 
     def predict(self, multiunits, time=None, is_compute_acausal=True,
-                use_gpu=False, state_names=None):
-        '''
+                use_gpu=False, state_names=None, store_likelihood=False):
+        """Run the state space model.
 
         Parameters
         ----------
         multiunits : array_like, shape (n_time, n_marks, n_electrodes)
-        time : None or ndarray, shape (n_time,)
+        time : None or np.ndarray, shape (n_time,)
         is_compute_acausal : bool, optional
             Use future information to compute the posterior.
         use_gpu : bool, optional
             Use GPU for the state space part of the model, not the likelihood.
         state_names : None or array_like, shape (n_states,)
+        store_likelihood : bool, optional
+            Save the likelihood for EM computations
 
         Returns
         -------
         results : xarray.Dataset
 
-        '''
+        """
         multiunits = np.asarray(multiunits)
         n_time = multiunits.shape[0]
 
@@ -743,6 +883,8 @@ class ClusterlessClassifier(_ClassifierBase):
                     is_track_interior=is_track_interior,
                     **encoding_params
             )
+        if store_likelihood:
+            self.likelihood_ = likelihood
 
         return self._get_results(
-            likelihood, n_time, time, state_names, use_gpu, is_compute_acausal)
+            likelihood, n_time, time, is_compute_acausal, use_gpu, state_names)
