@@ -1,3 +1,5 @@
+"""State space models that decode trajectories from population spiking"""
+
 from copy import deepcopy
 from logging import getLogger
 
@@ -5,21 +7,19 @@ import joblib
 import numpy as np
 import sklearn
 import xarray as xr
-from replay_trajectory_classification.bins import atleast_2d, get_centers
 from replay_trajectory_classification.continuous_state_transitions import (
     EmpiricalMovement, RandomWalk)
 from replay_trajectory_classification.core import (_acausal_decode,
                                                    _acausal_decode_gpu,
                                                    _causal_decode,
-                                                   _causal_decode_gpu, mask,
-                                                   scaled_likelihood)
+                                                   _causal_decode_gpu,
+                                                   atleast_2d, get_centers,
+                                                   mask, scaled_likelihood)
 from replay_trajectory_classification.environments import Environment
 from replay_trajectory_classification.initial_conditions import \
     UniformInitialConditions
 from replay_trajectory_classification.likelihoods import (
-    _ClUSTERLESS_ALGORITHMS, estimate_place_fields,
-    estimate_spiking_likelihood)
-from replay_trajectory_classification.misc import NumbaKDE
+    _SORTED_SPIKES_ALGORITHMS, _ClUSTERLESS_ALGORITHMS)
 from sklearn.base import BaseEstimator
 
 logger = getLogger(__name__)
@@ -27,10 +27,14 @@ logger = getLogger(__name__)
 sklearn.set_config(print_changed_only=False)
 
 _DEFAULT_CLUSTERLESS_MODEL_KWARGS = {
-    'model': NumbaKDE,
-    'model_kwargs': {
-        'bandwidth': np.array([24.0, 24.0, 24.0, 24.0, 6.0, 6.0])
-    }
+    'mark_std': 24.0,
+    'position_std': 6.0,
+}
+
+_DEFAULT_SORTED_SPIKES_MODEL_KWARGS = {
+    'position_std': 6.0,
+    'use_diffusion_distance': False,
+    'block_size': None,
 }
 
 
@@ -92,6 +96,63 @@ class _DecoderBase(BaseEstimator):
     def copy(self):
         return deepcopy(self)
 
+    def _get_results(
+        self,
+        results,
+        n_time,
+        time=None,
+        is_compute_acausal=True,
+        use_gpu=False,
+    ):
+
+        is_track_interior = self.environment.is_track_interior_.ravel(
+            order='F')
+        n_position_bins = is_track_interior.shape[0]
+        st_interior_ind = np.ix_(is_track_interior, is_track_interior)
+
+        logger.info('Estimating causal posterior...')
+        if not use_gpu:
+            results['causal_posterior'] = np.full(
+                (n_time, n_position_bins), np.nan, dtype=np.float64)
+            (results['causal_posterior'][:, is_track_interior],
+             data_log_likelihood) = _causal_decode(
+                self.initial_conditions_[is_track_interior].astype(np.float64),
+                self.state_transition_[st_interior_ind].astype(np.float64),
+                results['likelihood'][:, is_track_interior].astype(np.float64))
+        else:
+            results['causal_posterior'] = np.full(
+                (n_time, n_position_bins), np.nan, dtype=np.float32)
+            (results['causal_posterior'][:, is_track_interior],
+             data_log_likelihood) = _causal_decode_gpu(
+                self.initial_conditions_[is_track_interior],
+                self.state_transition_[st_interior_ind],
+                results['likelihood'][:, is_track_interior])
+
+        if is_compute_acausal:
+            logger.info('Estimating acausal posterior...')
+            if not use_gpu:
+                results['acausal_posterior'] = np.full(
+                    (n_time, n_position_bins, 1), np.nan, dtype=np.float64)
+                results['acausal_posterior'][:, is_track_interior] = (
+                    _acausal_decode(
+                        results['causal_posterior'][
+                            :, is_track_interior, np.newaxis].astype(np.float64),
+                        self.state_transition_[st_interior_ind].astype(np.float64)))
+            else:
+                results['acausal_posterior'] = np.full(
+                    (n_time, n_position_bins, 1), np.nan, dtype=np.float32)
+                results['acausal_posterior'][:, is_track_interior] = (
+                    _acausal_decode_gpu(
+                        results['causal_posterior'][
+                            :, is_track_interior, np.newaxis],
+                        self.state_transition_[st_interior_ind]))
+
+        if time is None:
+            time = np.arange(n_time)
+
+        return self.convert_results_to_xarray(
+            results, time, data_log_likelihood)
+
     def convert_results_to_xarray(self, results, time, data_log_likelihood):
         n_position_dims = self.environment.place_bin_centers_.shape[1]
         n_time = time.shape[0]
@@ -133,15 +194,16 @@ class _DecoderBase(BaseEstimator):
 
 
 class SortedSpikesDecoder(_DecoderBase):
-    def __init__(self,
-                 environment=Environment(environment_name=''),
-                 transition_type=RandomWalk(),
-                 initial_conditions_type=UniformInitialConditions(),
-                 infer_track_interior=True,
-                 knot_spacing=10,
-                 spike_model_penalty=1E1,
-                 ):
-        '''
+    def __init__(
+        self,
+        environment=Environment(environment_name=''),
+        transition_type=RandomWalk(),
+        initial_conditions_type=UniformInitialConditions(),
+        infer_track_interior=True,
+        sorted_spikes_algorithm='spiking_likelihood_kde',
+        sorted_spikes_algorithm_params=_DEFAULT_SORTED_SPIKES_MODEL_KWARGS,
+    ):
+        """
 
         Attributes
         ----------
@@ -157,29 +219,35 @@ class SortedSpikesDecoder(_DecoderBase):
             How far apart the spline knots are in position
         spike_model_penalty : float, optional
             L2 penalty (ridge) for the size of the regression coefficients
-        '''
+        """
         super().__init__(environment,
                          transition_type,
                          initial_conditions_type,
                          infer_track_interior)
-        self.knot_spacing = knot_spacing
-        self.spike_model_penalty = spike_model_penalty
+        self.sorted_spikes_algorithm = sorted_spikes_algorithm
+        self.sorted_spikes_algorithm_params = sorted_spikes_algorithm_params
 
     def fit_place_fields(self, position, spikes, is_training=None):
         logger.info('Fitting place fields...')
         if is_training is None:
             is_training = np.ones((position.shape[0],), dtype=np.bool)
         is_training = np.asarray(is_training).squeeze()
-        self.place_fields_ = estimate_place_fields(
-            position[is_training],
-            spikes[is_training],
-            self.environment.place_bin_centers_,
-            self.environment.place_bin_edges_,
-            penalty=self.spike_model_penalty,
-            knot_spacing=self.knot_spacing)
+        kwargs = self.sorted_spikes_algorithm_params
+        if kwargs is None:
+            kwargs = {}
+        self.place_fields_ = _SORTED_SPIKES_ALGORITHMS[
+            self.sorted_spikes_algorithm][0](
+                position[is_training],
+                spikes[is_training],
+                place_bin_centers=self.environment.place_bin_centers_,
+                place_bin_edges=self.environment.place_bin_edges_,
+                edges=self.environment.edges_,
+                is_track_interior=self.environment.is_track_interior_,
+                is_track_boundary=self.environment.is_track_boundary_,
+                **kwargs)
 
     def plot_place_fields(self, sampling_frequency=1, col_wrap=5):
-        '''Plots the fitted 2D place fields for each neuron.
+        """Plots the fitted 2D place fields for each neuron.
 
         Parameters
         ----------
@@ -190,33 +258,33 @@ class SortedSpikesDecoder(_DecoderBase):
         -------
         g : xr.plot.FacetGrid instance
 
-        '''
+        """
         try:
-            g = (self.place_fields_.unstack('position') * sampling_frequency
+            g = (self.place_fields_.unstack('position').where(self.environment.is_track_interior_) * sampling_frequency
                  ).plot(x='x_position', y='y_position', col='neuron',
-                        hue='encoding_group', col_wrap=col_wrap)
+                        col_wrap=col_wrap)
         except ValueError:
             g = (self.place_fields_ * sampling_frequency).plot(
-                x='position', col='neuron', hue='encoding_group',
+                x='position', col='neuron',
                 col_wrap=col_wrap)
 
         return g
 
     def fit(self, position, spikes, is_training=None):
-        '''
+        """
 
         Parameters
         ----------
-        position : ndarray, shape (n_time, n_position_dims)
-        spikes : ndarray, shape (n_time, n_neurons)
-        is_training : None or bool ndarray, shape (n_time), optional
+        position : np.ndarray, shape (n_time, n_position_dims)
+        spikes : np.ndarray, shape (n_time, n_neurons)
+        is_training : None or bool np.ndarray, shape (n_time), optional
             Time bins to be used for encoding.
 
         Returns
         -------
         self
 
-        '''
+        """
         position = atleast_2d(np.asarray(position))
         spikes = np.asarray(spikes)
         self.fit_environment(position)
@@ -229,11 +297,11 @@ class SortedSpikesDecoder(_DecoderBase):
 
     def predict(self, spikes, time=None, is_compute_acausal=True,
                 use_gpu=False):
-        '''
+        """
 
         Parameters
         ----------
-        spikes : ndarray, shape (n_time, n_neurons)
+        spikes : np.ndarray, shape (n_time, n_neurons)
         time : ndarray or None, shape (n_time,), optional
         is_compute_acausal : bool, optional
         use_gpu : bool, optional
@@ -243,62 +311,21 @@ class SortedSpikesDecoder(_DecoderBase):
         -------
         results : xarray.Dataset
 
-        '''
+        """
         spikes = np.asarray(spikes)
-        is_track_interior = self.environment.is_track_interior_.ravel(
-            order='F')
         n_time = spikes.shape[0]
-        n_position_bins = is_track_interior.shape[0]
-        st_interior_ind = np.ix_(is_track_interior, is_track_interior)
 
-        results = {}
         logger.info('Estimating likelihood...')
+        results = {}
         results['likelihood'] = scaled_likelihood(
-            estimate_spiking_likelihood(
+            _SORTED_SPIKES_ALGORITHMS[self.sorted_spikes_algorithm][1](
                 spikes, np.asarray(self.place_fields_)))
-
-        logger.info('Estimating causal posterior...')
-        results['causal_posterior'] = np.full(
-            (n_time, n_position_bins), np.nan)
-        if not use_gpu:
-            (results['causal_posterior'][:, is_track_interior],
-             log_data_likelihood) = _causal_decode(
-                self.initial_conditions_[is_track_interior],
-                self.state_transition_[st_interior_ind],
-                results['likelihood'][:, is_track_interior])
-        else:
-            (results['causal_posterior'][:, is_track_interior],
-             log_data_likelihood) = _causal_decode_gpu(
-                self.initial_conditions_[is_track_interior],
-                self.state_transition_[st_interior_ind],
-                results['likelihood'][:, is_track_interior])
-
-        if is_compute_acausal:
-            logger.info('Estimating acausal posterior...')
-            results['acausal_posterior'] = np.full(
-                (n_time, n_position_bins, 1), np.nan)
-            if not use_gpu:
-                results['acausal_posterior'][:, is_track_interior] = (
-                    _acausal_decode(
-                        results['causal_posterior'][
-                            :, is_track_interior, np.newaxis],
-                        self.state_transition_[st_interior_ind]))
-            else:
-                results['acausal_posterior'][:, is_track_interior] = (
-                    _acausal_decode_gpu(
-                        results['causal_posterior'][
-                            :, is_track_interior, np.newaxis],
-                        self.state_transition_[st_interior_ind]))
-
-        if time is None:
-            time = np.arange(n_time)
-
-        return self.convert_results_to_xarray(
-            results, time, log_data_likelihood)
+        return self._get_results(
+            results, n_time, time, is_compute_acausal, use_gpu)
 
 
 class ClusterlessDecoder(_DecoderBase):
-    '''
+    """
 
     Attributes
     ----------
@@ -314,7 +341,7 @@ class ClusterlessDecoder(_DecoderBase):
         The type of clusterless algorithm. See _ClUSTERLESS_ALGORITHMS for keys
     clusterless_algorithm_params : dict
         Parameters for the clusterless algorithms.
-    '''
+    """
 
     def __init__(self,
                  environment=Environment(environment_name=''),
@@ -332,7 +359,7 @@ class ClusterlessDecoder(_DecoderBase):
         self.clusterless_algorithm_params = clusterless_algorithm_params
 
     def fit_multiunits(self, position, multiunits, is_training=None):
-        '''
+        """
 
         Parameters
         ----------
@@ -340,7 +367,7 @@ class ClusterlessDecoder(_DecoderBase):
         multiunits : array_like, shape (n_time, n_marks, n_electrodes)
         is_training : None or array_like, shape (n_time,)
 
-        '''
+        """
         logger.info('Fitting multiunits...')
         if is_training is None:
             is_training = np.ones((position.shape[0],), dtype=np.bool)
@@ -361,7 +388,7 @@ class ClusterlessDecoder(_DecoderBase):
         )
 
     def fit(self, position, multiunits, is_training=None):
-        '''
+        """
 
         Parameters
         ----------
@@ -373,7 +400,7 @@ class ClusterlessDecoder(_DecoderBase):
         -------
         self
 
-        '''
+        """
         position = atleast_2d(np.asarray(position))
         multiunits = np.asarray(multiunits)
 
@@ -387,12 +414,12 @@ class ClusterlessDecoder(_DecoderBase):
 
     def predict(self, multiunits, time=None, is_compute_acausal=True,
                 use_gpu=False):
-        '''
+        """
 
         Parameters
         ----------
         multiunits : array_like, shape (n_time, n_marks, n_electrodes)
-        time : None or ndarray, shape (n_time,)
+        time : None or np.ndarray, shape (n_time,)
         is_compute_acausal : bool, optional
             Use future information to compute the posterior.
         use_gpu : bool, optional
@@ -402,16 +429,14 @@ class ClusterlessDecoder(_DecoderBase):
         -------
         results : xarray.Dataset
 
-        '''
+        """
         multiunits = np.asarray(multiunits)
         is_track_interior = self.environment.is_track_interior_.ravel(
             order='F')
         n_time = multiunits.shape[0]
-        n_position_bins = is_track_interior.shape[0]
-        st_interior_ind = np.ix_(is_track_interior, is_track_interior)
 
-        results = {}
         logger.info('Estimating likelihood...')
+        results = {}
         results['likelihood'] = scaled_likelihood(
             _ClUSTERLESS_ALGORITHMS[self.clusterless_algorithm][1](
                 multiunits=multiunits,
@@ -419,45 +444,6 @@ class ClusterlessDecoder(_DecoderBase):
                 is_track_interior=is_track_interior,
                 **self.encoding_model_
             ))
-        logger.info('Estimating causal posterior...')
-        if not use_gpu:
-            results['causal_posterior'] = np.full(
-                (n_time, n_position_bins), np.nan, dtype=np.float64)
-            (results['causal_posterior'][:, is_track_interior],
-             data_log_likelihood) = _causal_decode(
-                self.initial_conditions_[is_track_interior],
-                self.state_transition_[st_interior_ind],
-                results['likelihood'][:, is_track_interior])
-        else:
-            results['causal_posterior'] = np.full(
-                (n_time, n_position_bins), np.nan, dtype=np.float32)
-            (results['causal_posterior'][:, is_track_interior],
-             data_log_likelihood) = _causal_decode_gpu(
-                self.initial_conditions_[is_track_interior],
-                self.state_transition_[st_interior_ind],
-                results['likelihood'][:, is_track_interior])
 
-        if is_compute_acausal:
-            logger.info('Estimating acausal posterior...')
-            if not use_gpu:
-                results['acausal_posterior'] = np.full(
-                    (n_time, n_position_bins, 1), np.nan, dtype=np.float64)
-                results['acausal_posterior'][:, is_track_interior] = (
-                    _acausal_decode(
-                        results['causal_posterior'][
-                            :, is_track_interior, np.newaxis],
-                        self.state_transition_[st_interior_ind]))
-            else:
-                results['acausal_posterior'] = np.full(
-                    (n_time, n_position_bins, 1), np.nan, dtype=np.float32)
-                results['acausal_posterior'][:, is_track_interior] = (
-                    _acausal_decode_gpu(
-                        results['causal_posterior'][
-                            :, is_track_interior, np.newaxis],
-                        self.state_transition_[st_interior_ind]))
-
-        if time is None:
-            time = np.arange(n_time)
-
-        return self.convert_results_to_xarray(
-            results, time, data_log_likelihood)
+        return self._get_results(
+            results, n_time, time, is_compute_acausal, use_gpu)
