@@ -27,7 +27,7 @@ def make_simulated_run_data(sampling_frequency=SAMPLING_FREQUENCY,
                             place_field_variance=PLACE_FIELD_VARIANCE,
                             place_field_means=PLACE_FIELD_MEANS,
                             make_inbound_outbound_neurons=False):
-    '''Make simulated data of a rat running back and forth
+    """Make simulated data of a rat running back and forth
     on a linear maze with sorted spikes.
 
     Parameters
@@ -37,17 +37,17 @@ def make_simulated_run_data(sampling_frequency=SAMPLING_FREQUENCY,
     running_speed : float, optional
     n_runs : int, optional
     place_field_variance : float, optional
-    place_field_means : ndarray, shape (n_neurons,), optional
+    place_field_means : np.ndarray, shape (n_neurons,), optional
 
     Returns
     -------
-    time : ndarray, shape (n_time,)
-    linear_distance : ndarray, shape (n_time,)
+    time : np.ndarray, shape (n_time,)
+    linear_distance : np.ndarray, shape (n_time,)
     sampling_frequency : float
-    spikes : ndarray, shape (n_time, n_neurons)
-    place_fields : ndarray, shape (n_time, n_neurons)
+    spikes : np.ndarray, shape (n_time, n_neurons)
+    place_fields : np.ndarray, shape (n_time, n_neurons)
 
-    '''
+    """
     n_samples = int(n_runs * sampling_frequency *
                     2 * track_height / running_speed)
 
@@ -99,7 +99,8 @@ def make_continuous_replay(sampling_frequency=SAMPLING_FREQUENCY,
                            is_outbound=True):
 
     replay_speed = running_speed * replay_speedup
-    n_samples = int(2 * sampling_frequency * track_height / replay_speed)
+    n_samples = int(
+        np.ceil(2 * sampling_frequency * track_height / replay_speed))
     replay_time = simulate_time(n_samples, sampling_frequency)
     true_replay_position = simulate_linear_distance(
         replay_time, track_height, replay_speed)
@@ -188,3 +189,17 @@ def make_fragmented_continuous_fragmented_replay(
     replay_time = np.arange(test_spikes.shape[0]) / sampling_frequency
 
     return replay_time, test_spikes
+
+
+def make_theta_sweep(sampling_frequency=SAMPLING_FREQUENCY, n_sweeps=5):
+    _, test_spikes1 = make_continuous_replay(
+        is_outbound=False, replay_speedup=145)
+    _, test_spikes2 = make_continuous_replay(
+        is_outbound=True, replay_speedup=145)
+
+    test_spikes = np.concatenate(
+        [test_spikes1[test_spikes1.shape[0] // 2:],
+         test_spikes2] * n_sweeps)
+    replay_time = np.arange(test_spikes.shape[0]) / sampling_frequency
+
+    return replay_time, test_spikes
